/* ------------------------------------------------------------------------*/
/* COPYRIGHT (c) 1992, 1995, 1996, 1998 CONTEMPORARY CONTROL SYSTEMS, INC. */
/* ------------------------------------------------------------------------*/
/* Project:       20020 DRIVER                                             */
/* Filename:      arc20.C                                                  */
/* Description:   MAIN ROUTINES TO USE ANY COM20020 BASED BOARD            */
/* Version:       2.20                                                     */
/* Authors:       Mike Justice, Bennet Levine                              */
/* History:       ORIGINAL 09/14/92                                        */
/*                11/22/95 Improved Interrupt handling                     */
/*                02/03/95 fixed problem w/writing command chaining        */
/*                04/23/96 added support for long packets                  */  
/*                09/14/98 save/restore address pointers in ISR            */
/*                         provide proper arc20_write_abort operation      */
/* ------------------------------------------------------------------------*/
/*                                                                         */
/*       arc20_init                 initialize the hardware                */
/*       arc20_read                 read a packet                          */
/*       arc20_write                write a packet                         */
/*       arc20_status               status of node                         */
/*       arc20_exit                 reset and cleanup                      */
/* ------------------------------------------------------------------------*/
#include <dos.h>
#include <stdio.h>
#include <conio.h>
#include "arcdef.h"
#include "msc.h"

/* recv,tx page defines */
#define ENABLE_PAGE0   0x84
#define ENABLE_PAGE1   0x8c


/* ------------------------------------------------------------------------- */
/* external variable definitions                                             */
/* ------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* global variable definition                                                */
/* ------------------------------------------------------------------------- */
USIGN8  arc20_nodeid;
USIGN16 arc20_port = 0x2e0;
USIGN16 arc20_swport = 0x2e9;

/* interrupt mode = TRUE, polled mode = FALSE */
USIGN8  arc20_intr = TRUE;
USIGN8  arc20_level = 2;

/* ------------------------------------------------------------------------- */
/* local variable definition                                                 */
/* ------------------------------------------------------------------------- */
static  USIGN8  i,j,k;
static  USIGN8 count[2];

USIGN8  arc20_stat;
USIGN8  arc20_page;
USIGN8  arc20_offset;
USIGN8  arc20_nbytes;
USIGN8  arc20_config;
USIGN8  arc20_diag;
USIGN8  arc20_in_page;
USIGN8  arc20_out_page;
USIGN8  arc20_cmd;
USIGN8  arc20_rx_flag;
USIGN8  arc20_write_ack;
USIGN8  arc20_write_abort;
USIGN8  arc20_buffer[512];

USIGN16 arc20_interrupt_reg;
USIGN16 arc20_status_reg;
USIGN16 arc20_command_reg;
USIGN16 arc20_diag_reg;
USIGN16 arc20_addr_hi_reg;
USIGN16 arc20_addr_low_reg;
USIGN16 arc20_packet_mem;
USIGN16 arc20_config_reg;
USIGN16 arc20_tentid_reg;
USIGN16 arc20_nextid_reg;
USIGN16 arc20_nodeid_reg;
USIGN16 arc20_setup_reg;

/* interrupt vector definitions */
#define INT0  (USIGN32 far *)   0x20L
USIGN32 far *oldvector;
USIGN32 far *farptr;
USIGN8  mask;

/* ------------------------------------------------------------------------- */
/* function prototypes */
/* ------------------------------------------------------------------------- */
#include "arc20.h"


/****************************************************************************
 * FUNCTION NAME: arc20_init
 *
 * DESCRIPTION  : initial the node and join the network
 *
 * RETURN VALUE : status
 ***************************************************************************/
USIGN8 arc20_init()
{
    USIGN8 status;

    if (arc20_intr == TRUE) printf("Initializing 20020 at %XH, Interrupt Level %d\n",arc20_port,arc20_level);
    else printf("Initializing 20020 at %XH, No interrupts\n",arc20_port);

    /* an 8 bit interface uses +1x, a 16 bit interface users +2x */
    arc20_interrupt_reg = arc20_port;
    arc20_status_reg = arc20_port;
    arc20_command_reg = arc20_port + 1;
    arc20_diag_reg = arc20_port + 1;
    arc20_addr_hi_reg = arc20_port + 2;
    arc20_addr_low_reg = arc20_port + 3;
    arc20_packet_mem = arc20_port + 4;
    arc20_config_reg = arc20_port + 6;
    arc20_tentid_reg = arc20_port + 7;
    arc20_nextid_reg = arc20_port + 7;
    arc20_nodeid_reg = arc20_port + 7;
    arc20_setup_reg = arc20_port + 7;

    /* set the page numbers */
    arc20_in_page = 0;
    arc20_out_page = 2;
    arc20_rx_flag = FALSE;

    /* setup to use node id register */
    arc20_config = (USIGN8)INP(arc20_config_reg);
    arc20_config = (USIGN8)((arc20_config & SUB_AD) | NODE_ID_REG);
    OUTP(arc20_config_reg, arc20_config);

    /* clear the POR and RECON status bits */
    arc20_cmd = CLEARFLAGS;
    OUTP(arc20_command_reg, arc20_cmd);

    /* determine the node id - read DIP switch */
    arc20_nodeid = (USIGN8)INP(arc20_swport);
    if (arc20_nodeid == 0) printf("Invalid Nodeid\n");
    OUTP(arc20_nodeid_reg, arc20_nodeid);    
    DELAY(200);    

    /* reset the 20020 chip - toggle it */
    arc20_config = (USIGN8)INP(arc20_config_reg);
    arc20_config = (USIGN8)(arc20_config | RESET);
    OUTP(arc20_config_reg, arc20_config);
    arc20_config = (USIGN8)(arc20_config & ~RESET);
    OUTP(arc20_config_reg, arc20_config);
    DELAY(200);    

    /* get the reset indicator it must be a 0xd1 */
    OUTP(arc20_addr_hi_reg,0xc0);
    OUTP(arc20_addr_low_reg,0);
    arc20_buffer[0] = (USIGN8)INP(arc20_packet_mem);
    arc20_buffer[1] = (USIGN8)INP(arc20_packet_mem);
    if (arc20_buffer[0] != 0xD1) status = 0;
    else {
        /* mask interrupts */
        OUTP(arc20_interrupt_reg,0);

        /* disable tx,rx */
        OUTP(arc20_command_reg, DISABLETX);
        OUTP(arc20_command_reg, DISABLERX);

        /* clear the POR and RECON status bits */
        arc20_cmd = CLEARFLAGS;
        OUTP(arc20_command_reg, arc20_cmd);

        /* short or long packets */
        OUTP(arc20_command_reg, LONGPACKET);

        /* setup network speed - 2.5Mbps */
        arc20_config = (USIGN8)((arc20_config & SUB_AD) | SETUP_REG);
        OUTP(arc20_config_reg, arc20_config);
        OUTP(arc20_setup_reg, 0);

        /* interrupt driven */
        if (arc20_intr == TRUE) 
        {
            /* exchange vectors based on PC hardware and 8259 */
            DISABLE;
            farptr = INT0;
            for (i=0; i<arc20_level; i++) farptr++;
            *oldvector = *farptr;
            *farptr = (USIGN32) (VOID far *) arc20_interrupt;
            mask = (USIGN8)INP(0x21);
            mask &= ~(1 << arc20_level);
            OUTP(0x21, mask);

            /* enable recv, recon, exenak */
            OUTP(arc20_interrupt_reg, 0x8c);
            ENABLE;
        }

        /* configuration - network timeout slowest, enable tx */
        arc20_config = US78 | TXEN | CCHEN;

        /* output the configuration - join the network */
        OUTP(arc20_config_reg, arc20_config);

        /* enable two reads, w/broadcast */
        arc20_cmd = ENABLE_PAGE0;
        OUTP(arc20_command_reg,arc20_cmd);
        arc20_cmd = ENABLE_PAGE1;
        OUTP(arc20_command_reg,arc20_cmd);

        /* return the node ID */
        arc20_nodeid = arc20_buffer[1];
        status = arc20_nodeid;
    }

    /* return 0 if bad initialization, or node ID */
    return(status);
}

/****************************************************************************
 * FUNCTION NAME: arc20_read
 *
 * DESCRIPTION  : read a packet
 *
 * RETURN VALUE : status
 ***************************************************************************/
USIGN8 arc20_read(USIGN8 *data_ptr)
{
    USIGN8 status;
    USIGN16 bufcount,i;

    /* check if interrupt receive */
    if (arc20_intr == TRUE)
    {
        /* check if buffers have been filled */
        if (arc20_rx_flag == TRUE)
        {
            /* get the buffer count */
            bufcount = arc20_buffer[3] + 4;
            if (arc20_buffer[2] > 0) bufcount += 256;

            /* transfer the buffer to user */
            for (i=0; i<bufcount; i++)
            {
                *(data_ptr+i) = arc20_buffer[i];
            }         
            
            arc20_rx_flag = FALSE;
            status = E_OK;
        }
        else
        {
            /* no packets */
            status = E_NO_PACKET;
        }
    }
    else
    {
        /* check the recieve (RI) bit */
        arc20_get_check_status();
        status = arc20_stat;
        if ((status & RI) == RI) 
        {
            /* reset command chaining */
            OUTP(arc20_command_reg,0x08);

            /* enable the next read page */
            arc20_page = arc20_in_page;
            arc20_in_page++;
            arc20_in_page &= 1;
            status = E_OK;

            /* read the 1st 2 bytes sid,did */
            if (arc20_page == 0)
            {
                OUTP(arc20_addr_hi_reg,0xc0);
            }
            else
            {
                OUTP(arc20_addr_hi_reg,0xc2);
            }
            OUTP(arc20_addr_low_reg,0);
            *data_ptr = (USIGN8)INP(arc20_packet_mem);
            *(data_ptr+1) = (USIGN8)INP(arc20_packet_mem);

            /* read the packet size */
            count[0] = (USIGN8)INP(arc20_packet_mem);
            count[1] = (USIGN8)INP(arc20_packet_mem);

            /* store real count */

            /* long or short */
            if (count[0] == 0) 
            {
                arc20_offset = count[1];
                arc20_nbytes = (USIGN8)((~count[1])+1);
                *(data_ptr+3) = arc20_nbytes;
                *(data_ptr+2) = 1;
                bufcount = 256 + arc20_nbytes;
            }
            else
            {
                arc20_offset = count[0];
                arc20_nbytes = (USIGN8)((~count[0])+1);
                *(data_ptr+3) = arc20_nbytes;
                *(data_ptr+2) = 0;
                bufcount = arc20_nbytes;
            }

            /* read the next bytes */
            if (arc20_page == 0)
            {
                OUTP(arc20_addr_hi_reg,0xc0);
            }
            else
            {
                OUTP(arc20_addr_hi_reg,0xc2);
            }
            OUTP(arc20_addr_low_reg,arc20_offset);
            
            /* transfer the buffer to storage */
            for (i=0; i<bufcount; i++)
            {
                *(data_ptr+4+i) = (USIGN8)INP(arc20_packet_mem);
            }

            /* enable this page */
            if (arc20_in_page == 0) arc20_cmd = ENABLE_PAGE1;
            else arc20_cmd = ENABLE_PAGE0;
            OUTP(arc20_command_reg,arc20_cmd);
        }
        else
        {
            status = E_NO_PACKET;
        }
    }

    /* return system status */
    return(status);
}

/****************************************************************************
 * FUNCTION NAME: arc20_write
 *
 * DESCRIPTION  : write a packet
 *
 * RETURN VALUE : status
 ***************************************************************************/
USIGN8 arc20_write(USIGN8 *data_ptr)
{
    USIGN8 page;
    USIGN8 count;
    USIGN8 lors;
    USIGN8 offset;
    USIGN8 status;
    USIGN8 counts[2];

    /* check if available */
    arc20_get_check_status();
    status = arc20_stat;
    if ((status & TA2) > 0) 
    {
        status = E_OK;
        page = arc20_out_page;

        /* SID and DID */
        arc20_write_data(page, 0, 2, 0, data_ptr);

        /* user packet format - SID, DID, msbcount, lsbcount, data,... */
        data_ptr += 2;

        /* determine the count and packet size to use */
        if (*data_ptr > 0) {

            /* long packet */
            lors = 1;
            data_ptr++;
            count = *data_ptr;
            if (count == 0) status = E_BAD_PACKET_SIZE;
            else {
                offset = (USIGN8)((~count)+1);
                counts[0] = 0;
                counts[1] = offset;
            }
        }
        else {
            /* short packet */
            lors = 0;
            data_ptr++;
            count = *data_ptr;
            if (count > 253) status = E_BAD_PACKET_SIZE;
            else {
                offset = (USIGN8)((~count)+1);
                counts[0] = offset;
            }
        } 

        /* load the buffer */
        if (status == E_OK) {

            /* load the count/fp */
            arc20_write_data(page, 2, 2, 0, &counts[0]);

            /* load the data */
            data_ptr++;
            arc20_write_data(page, offset, count, lors, data_ptr);

            /* reset ack flag and abort flag */
            arc20_write_ack = FALSE;
            arc20_write_abort = FALSE;

            /* send it out */
            page <<= 3;
            page |= 3;
            OUTP(arc20_command_reg, page);

            /* incr page and check for rollover */
            arc20_out_page++;
            if (arc20_out_page > 3) arc20_out_page = 2;
        }
    }
    else status = E_TX_BUSY;

    /* return system status */
    return(status);
}

/****************************************************************************
 * FUNCTION NAME: arc20_status
 *
 * DESCRIPTION  : status of network
 *
 * RETURN VALUE : status
 ***************************************************************************/
USIGN8 arc20_status(void)
{
    /* get the status */
    arc20_get_check_status();

    /* return status */
    return(arc20_stat);
}


/****************************************************************************
 * FUNCTION NAME: arc20_interrupt
 *
 * DESCRIPTION  : interrupt routine
 *
 * RETURN VALUE : none
 ***************************************************************************/
void interrupt far arc20_interrupt(void)
{
    USIGN16 i,bytes;
    USIGN8 addr_low_save, addr_high_save;

    /* get the status */
    arc20_stat = (USIGN8)INP(arc20_status_reg);
    arc20_diag = (USIGN8)INP(arc20_diag_reg);

    /* receiver */
    if ((arc20_stat & RI) == RI)
    {
        /* set the flag */
        arc20_rx_flag = TRUE;

        /* reset command chaining */
        OUTP(arc20_command_reg,0x08);

        /* get the current page */
        arc20_page = arc20_in_page;
        arc20_in_page++;
        arc20_in_page &= 1;

        /* save the address_low register and the address_high register */
        /* before modification */
        addr_low_save = INP(arc20_addr_low_reg);
        addr_high_save  = INP(arc20_addr_hi_reg);

        /* read the 1st 2 bytes sid,did */
        if (arc20_page == 0)
        {
            OUTP(arc20_addr_hi_reg,0xc0);
        }
        else
        {
            OUTP(arc20_addr_hi_reg,0xc2);
        }
        OUTP(arc20_addr_low_reg,0);
        arc20_buffer[0] = (USIGN8)INP(arc20_packet_mem);
        arc20_buffer[1] = (USIGN8)INP(arc20_packet_mem);

        /* read the packet size */
        count[0] = (USIGN8)INP(arc20_packet_mem);
        count[1] = (USIGN8)INP(arc20_packet_mem);

        /* store real count */
        /* long or short */
        if (count[0] == 0) 
        {
            arc20_offset = count[1];
            arc20_nbytes = (USIGN8)((~count[1])+1);
            arc20_buffer[3] = arc20_nbytes;
            arc20_buffer[2] = 1;
            bytes = 256 + arc20_nbytes;
        }
        else
        {
            arc20_offset = count[0];
            arc20_nbytes = (USIGN8)((~count[0])+1);
            arc20_buffer[3] = arc20_nbytes;
            arc20_buffer[2] = 0;
            bytes = arc20_nbytes;
        }

        /* read the next bytes */
        if (arc20_page == 0)
        {
            OUTP(arc20_addr_hi_reg,0xc0);
        }
        else
        {
            OUTP(arc20_addr_hi_reg,0xc2);
        }
        OUTP(arc20_addr_low_reg,arc20_offset);
            
        /* transfer the buffer to storage */
        for (i=0; i<bytes; i++)
        {
            arc20_buffer[4+i] = (USIGN8)INP(arc20_packet_mem);
        }

        /* enable this page */
        if (arc20_in_page == 0) arc20_cmd = ENABLE_PAGE1;
        else arc20_cmd = ENABLE_PAGE0;
        OUTP(arc20_command_reg,arc20_cmd);

        /* restore original address_high and address_low registers */
        OUTP(arc20_addr_hi_reg, addr_high_save);
        OUTP(arc20_addr_low_reg, addr_low_save);
    }

    /* recon */
    if ((arc20_stat & RECON) > 0)
    {
        OUTP(arc20_command_reg,0x1e);
    }

    /* execessive nak */
    if ((arc20_diag & EXC_NAK) > 0)
    {
        /* let the write status routine abort the packet */
        OUTP(arc20_command_reg,0x1e);
        OUTP(arc20_command_reg,0x01);
        arc20_write_abort = TRUE;
    }

    /* send eoi */
    OUTP(0x20, 0x20);
}


/****************************************************************************
 * FUNCTION NAME: arc20_write_data
 *
 * DESCRIPTION  : Write data from buffer location to a specific page
 *
 * RETURN VALUE : none
 ***************************************************************************/
void arc20_write_data(USIGN8 page, USIGN8 offset, USIGN8 count, USIGN8 shortlong, 
                USIGN8 *user_buffer)
{
    USIGN16 address;

    /* set the address pointers, autoincrement on write */
    address = (512 * page) + offset;
    OUTP(arc20_addr_hi_reg, (USIGN8)((address >> 8) | AUTOINC) & WRDATA);
    OUTP(arc20_addr_low_reg, (USIGN8)address);

    /* write the bytes to the page */
    while(count--) 
    {
        OUTP(arc20_packet_mem, *user_buffer++);
    }

    /* do it another 256x */
    if (shortlong == 1) {
        count = 0;
        do {
            OUTP(arc20_packet_mem, *user_buffer);
            user_buffer++;
            count--;
        } while (count != 0);
    }
}


/****************************************************************************
 * FUNCTION NAME: arc20_get_check_status
 *
 * DESCRIPTION  : get and check the device status
 *
 * RETURN VALUE : none
 ***************************************************************************/
void arc20_get_check_status(void)
{
    /* read the status */
    arc20_stat = (USIGN8)INP(arc20_status_reg);
    arc20_diag = (USIGN8)INP(arc20_diag_reg);

    /* check for TMA ack */
    if ((arc20_stat & TMA) == TMA) {
        arc20_write_ack = TRUE; 
    }

    /* check for MYRECON */
    if ((arc20_diag & MYRECON) == MYRECON) {

        /* clear flags */
        OUTP(arc20_command_reg, CLEARFLAGS);
        arc20_stat = (USIGN8)INP(arc20_status_reg);
    }

    /* check for RECON */
    if ((arc20_stat & RECON) == RECON) {

        /* clear flags */
        OUTP(arc20_command_reg, CLEARFLAGS);
        arc20_stat = (USIGN8)INP(arc20_status_reg);
    }

    /* check for POR flag */
    if ((arc20_stat & POR) == POR) {

        /* clear flags */
        OUTP(arc20_command_reg, CLEARPOR);
    }

    /* check for EXENAK */
    if ((arc20_diag & EXC_NAK) == EXC_NAK) {

        /* clear flags */
        OUTP(arc20_command_reg, CLEARPOR);

        /* kill the last transmit */
        OUTP(arc20_command_reg, DISABLETX);

        /* notify the write routine */
        arc20_write_abort = TRUE;
    }

    /* read the status again */
    arc20_stat = (USIGN8)INP(arc20_status_reg);
    arc20_diag = (USIGN8)INP(arc20_diag_reg);
}


/****************************************************************************
 * FUNCTION NAME: arc20_exit  
 *
 * DESCRIPTION  : turn off 20020 and disable interrupt if used
 *
 * RETURN VALUE : none
 ***************************************************************************/
VOID arc20_exit(void)
{
    /* disable the transmitter */
    OUTP(arc20_command_reg, DISABLETX);

    /* disable the reciever */
    OUTP(arc20_command_reg, DISABLERX);

    /* disable the tx enable */
    arc20_config &= (USIGN8)~TXEN;
    OUTP(arc20_config_reg, arc20_config);

    /* reset the interrupt vector if set */
    if (arc20_intr == TRUE) {

        /* exchange vectors */
        DISABLE;
        farptr = INT0;
        for (i=0; i<arc20_level; i++) farptr++;
        *farptr = *oldvector;
        mask = (USIGN8)INP(0x21);
        mask |= (1 << arc20_level);
        OUTP(0x21, mask);
        OUTP(arc20_interrupt_reg, 0x00);
        ENABLE;
    }
}

